import argparse
import re

if re.match("^pyp2p", __name__) != None:
    # Setup parser.
    parser = argparse.ArgumentParser(prog='pyp2p')

    # Option: LAN IP.
    parser.add_argument('-lan_ip', '--lan_ip', action="store", dest="lan_ip", help="query string", default=None)

    # Option: LAN IP.
    parser.add_argument('-wan_ip', '--wan_ip', action="store", dest="wan_ip", help="query string", default=None)

    # Parse arguments.
    args = args, unknown = parser.parse_known_args()
else:
    class Args():
        def __init__(self):
            self.wan_ip = None
            self.lan_ip = None

    args = Args()



"""
The purpose of this server is to maintain a list of
nodes connected to the P2P network. It supports traditional
passive nodes which receive connections but it also
provides a protocol to facilitate TCP hole punching
allowing >85% of nodes behind a standard router to
receive connections.

Notes: Passives nodes don't keep a persistent connection
to the rendezvous server because they don't need to
and it would be unnecessarily taxing. For this reason,
the list of passive nodes retrieved from bootstrapping
may be stale but they are periodically cleaned every node
life time. Additionally, well behaved passive nodes
send the clear command which causes the server to remove
the node details.
"""

import re
import socket
import random
import time
import ipaddress
import os
import sys
from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor
from .lib import *

error_log_path = "error.log"
debug = 1

class RendezvousProtocol(LineReceiver):
    def __init__(self, factory):
        self.factory = factory
        self.challege_timeout = 60 * 2 # Seconds.
        self.node_lifetime = 60 * 60 * 12 # 12 hours.
        self.cleanup = 5 * 60 # Every 5 minutes.
        self.max_candidates = 100 # Per simultaneous node.
        self.connected = False

    def log_entry(self, msg, direction="none"):
        if sys.version_info >= (3,0,0):
            if type(msg) == bytes:
                msg = msg.decode("utf-8")
        ip_addr = str(self.transport.getPeer().host)
        port = str(self.transport.getPeer().port)
        when = time.strftime("%H:%M:%S %Y-%m-%d")
        who = """%s:%s""" % (ip_addr, port)
        if direction == "send":
            direction = " -> "
        elif direction == "recv":
            direction = " <- "
        else:
            direction = " "

        entry = """[%s] %s%s%s""" % (when, msg, direction, who)
        return entry

    def send_line(self, msg):
        # Not connected.
        if not self.connected:
            return

        # Network byte order.
        try:
            if type(msg) != bytes:
                msg = msg.encode("ascii")
        except Exception as e:
            print("send line e")
            print(e)
            return

        # stdout for debugging.
        if debug:
            print(self.log_entry(msg, "send"))

        self.sendLine(msg)
        

    def send_remote_port(self):
        """
        Sends the remote port mapped for the connection.
        This port is surprisingly often the same as the locally
        bound port for an endpoint because a lot of NAT types
        preserve the port.
        """
        msg = "REMOTE TCP %s" % (str(self.transport.getPeer().port))
        self.send_line(msg)

    def is_valid_ipv4_address(self, address):
        try:
            socket.inet_pton(socket.AF_INET, address)
        except AttributeError:  # no inet_pton here, sorry
            try:
                socket.inet_aton(address)
            except socket.error:
                return False
            return address.count('.') == 3
        except socket.error:  # not a valid address
            return False

        return True

    def is_valid_port(self, port):
        port = re.findall("^[0-9]+$", str(port))
        if len(port):
            port = int(port[0])
            if 0 < port <= 65535:
                return 1
        return 0

    def cleanup_candidates(self, node_ip):
        """
        Removes old TCP hole punching candidates for a
        designated node if a certain amount of time has passed
        since they last connected.
        """
        if node_ip in self.factory.candidates:
            old_candidates = []
            for candidate in self.factory.candidates[node_ip]:
                elapsed = int(time.time() - candidate["time"])
                if elapsed > self.challege_timeout:
                    old_candidates.append(candidate)

            for candidate in old_candidates:
                self.factory.candidates[node_ip].remove(candidate)

    def propogate_candidates(self, node_ip):
        """
        Used to progate new candidates to passive simultaneous
        nodes.
        """

        if node_ip in self.factory.candidates:
            old_candidates = []
            for candidate in self.factory.candidates[node_ip]:
                # Not connected.
                if not candidate["con"].connected:
                    continue

                # Already sent -- updated when they accept this challenge.
                if candidate["propogated"]:
                    continue

                # Notify node of challege from client.
                msg = "CHALLENGE %s %s %s" % (candidate["ip_addr"], " ".join(map(str, candidate["predictions"])), candidate["proto"])
                
                self.factory.nodes["simultaneous"][node_ip]["con"].send_line(msg)
                old_candidates.append(candidate)

    def synchronize_simultaneous(self, node_ip):
        """
        Because adjacent mappings for certain NAT types can
        be stolen by other connections, the purpose of this
        function is to ensure the last connection by a passive
        simultaneous node is recent compared to the time for
        a candidate to increase the chance that the precited
        mappings remain active for the TCP hole punching
        attempt.
        """

        for candidate in self.factory.candidates[node_ip]:
            # Only if candidate is connected.
            if not candidate["con"].connected:
                continue

            # Synchronise simultaneous node.
            if candidate["time"] - self.factory.nodes["simultaneous"][node_ip]["time"] > self.challege_timeout:
                msg = "RECONNECT"
                self.factory.nodes["simultaneous"][node_ip]["con"].send_line(msg)
                return
 
        self.cleanup_candidates(node_ip)
        self.propogate_candidates(node_ip)

    def connectionMade(self):
        try:
            self.connected = True
            if debug:
                print(self.log_entry("OPENED =", "none"))

            # Force reconnect if node has candidates and the timeout is old.
            ip_addr = self.transport.getPeer().host
            if ip_addr in self.factory.nodes["simultaneous"]:
                # Update time.
                self.factory.nodes["simultaneous"][ip_addr]["time"] = time.time()
                self.synchronize_simultaneous(ip_addr)
        except Exception as e:
            error = parse_exception(e)
            log_exception(error_log_path, error)
            print(self.log_entry("ERROR =", error))

    def connectionLost(self, reason):
        """
        Mostly handles clean-up of node + candidate structures.
        Avoids memory exhaustion for a large number of connections.
        """
        try:
            self.connected = False
            if debug:
                print(self.log_entry("CLOSED =", "none"))

            # Every five minutes: cleanup
            t = time.time()
            if time.time() - self.factory.last_cleanup >= self.cleanup:
                self.factory.last_cleanup = t

                # Delete old passive nodes.
                old_node_ips = []
                for node_ip in list(self.factory.nodes["passive"]):
                    passive_node = self.factory.nodes["passive"][node_ip]
                    # Gives enough time for passive nodes to receive clients.
                    if t - passive_node["time"] >= self.node_lifetime:
                        old_node_ips.append(node_ip)
                for node_ip in old_node_ips:
                    del self.factory.nodes["passive"][node_ip]

                # Delete old simultaneous nodes.
                old_node_ips = []
                for node_ip in list(self.factory.nodes["simultaneous"]):
                    simultaneous_node = self.factory.nodes["simultaneous"][node_ip]
                    # Gives enough time for passive nodes to receive clients.
                    if t - simultaneous_node["time"] >= self.node_lifetime:
                        old_node_ips.append(node_ip)
                for node_ip in old_node_ips:
                    del self.factory.nodes["simultaneous"][node_ip]

                # Delete old candidates and candidate structs.
                old_node_ips = []
                for node_ip in list(self.factory.candidates):
                    # Record old candidates.
                    old_candidates = []
                    for candidate in self.factory.candidates[node_ip]:
                        # Hole punching is ms time sensitive.
                        # Candidates older than this is safe to assume they're not needed.
                        if not node_ip in self.factory.nodes["simultaneous"] and t - candidate["time"] >= self.challenge_timeout * 5:
                            old_candidates.append(candidate)

                    # Remove old candidates.
                    for candidate in old_candidates:
                        self.factory.candidates[node_ip].remove(candidate)

                    # Record old node IPs.
                    if not len(self.factory.candidates[node_ip]) and not node_ip in self.factory.nodes["simultaneous"]:
                        old_node_ips.append(node_ip)

                # Remove old node IPs.
                for node_ip in old_node_ips:
                    del self.factory.candidates[node_ip]
        except Exception as e:
            error = parse_exception(e)
            log_exception(error_log_path, error)
            print(self.log_entry("ERROR =", error))

    def lineReceived(self, line):
        # Unicode for text patterns.
        try:
            line = line.decode("utf-8")
        except:
            # Received invalid characters.
            return
        if debug:
            print(self.log_entry(line, "recv"))

        try:
            # Return nodes for bootstrapping.
            if re.match("^BOOTSTRAP", line) is not None:
                parts = re.findall("^BOOTSTRAP ([0-9]+)", line)
                while 1:
                    # Invalid response.
                    if not len(parts):
                        break
                    n = int(parts[0])
            
                    # Invalid number.
                    if n < 1 or n > 100:
                        break

                    # Bootstrap n passive, n .
                    msg = "NODES "
                    node_types = ["passive"]
                    our_ip = self.transport.getPeer().host
                    node_no = 0
                    for node_type in node_types:
                        ip_addr_list = list(self.factory.nodes[node_type])
                        for i in range(0, n):
                            # There's no nodes left to bootstrap with.
                            ip_addr_list_len = len(ip_addr_list)
                            if not ip_addr_list_len:
                                break

                            # Choose a random node.
                            rand_index = random.randrange(0, ip_addr_list_len)
                            ip_addr = ip_addr_list[rand_index]
                            element = self.factory.nodes[node_type][ip_addr]

                            # Skip our own IP.
                            if our_ip == ip_addr or ip_addr == "127.0.0.1":
                                i -= 1
                                ip_addr_list.remove(ip_addr)
                                continue

                            # Not connected.
                            if node_type == "simultaneous" and not element["con"].connected:
                                i -= 1
                                ip_addr_list.remove(ip_addr)
                                continue

                            # Append new node.
                            msg += node_type[0] + ":" + ip_addr + ":" + str(element["port"]) + " "
                            ip_addr_list.remove(ip_addr)
                            node_no += 1

                    # No nodes in response.
                    if not node_no:
                        msg = "NODES EMPTY"

                    # Send nodes list.
                    self.send_line(msg)
                    break

            # Add node details to relevant sections.
            if re.match("^(SIMULTANEOUS|PASSIVE) READY [0-9]+ [0-9]+$", line) is not None:
                # Get type.
                node_type, passive_port, max_inbound = re.findall("^(SIMULTANEOUS|PASSIVE) READY ([0-9]+) ([0-9]+)", line)[0]
                node_type = node_type.lower()
                valid_node_types = ["simultaneous", "passive"]
                if node_type not in valid_node_types:
                    return

                # Init / setup.
                node_ip = self.transport.getPeer().host
                self.factory.nodes[node_type][node_ip] = {
                    "max_inbound": max_inbound,
                    "no": 0,
                    "port": passive_port,
                    "time": time.time(),
                    "con": self,
                    "ip_list": []
                }

                # Passive doesn't have a candidates list.
                if node_type == "simultaneous":
                    if not node_ip in self.factory.candidates:
                        self.factory.candidates[node_ip] = []
                    else:
                        self.cleanup_candidates(node_ip)
                        self.propogate_candidates(node_ip)

            # Echo back mapped port.
            if re.match("^SOURCE TCP", line) is not None:
                self.send_remote_port()

            # Client wishes to actively initate a simultaneous open.
            if re.match("^CANDIDATE", line) is not None:
                # CANDIDATE 192.168.0.1 TCP.
                parts = re.findall("^CANDIDATE ([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+) (TCP|UDP) ((?:[0-9]+\s?)+)$", line)
                while 1:
                    # Invalid response.
                    if not len(parts):
                        break
                    node_ip, proto, predictions = parts[0]
                    predictions = predictions.split(" ")
                    client_ip = self.transport.getPeer().host

                    # Invalid IP address.
                    if not self.is_valid_ipv4_address(node_ip):
                        print("Candidate invalid ip4" + str(node_ip))
                        break
                    if node_ip not in self.factory.nodes["simultaneous"]:
                        print("Candidate: node ip not in factory nodes sim.")
                        break
                    if node_ip == client_ip:
                        print("Candidate node ip == clietn ip")
                        break

                    # Valid port.
                    valid_ports = 1
                    for port in predictions:
                        if not self.is_valid_port(port):
                            valid_ports = 0
                    if not valid_ports:
                        print("Candidate not valid port")
                        break

                    # Not connected.
                    if not self.factory.nodes["simultaneous"][node_ip]["con"].connected:
                        print("Candidate not connected.")
                        break

                    candidate = {
                        "ip_addr": client_ip,
                        "time": time.time(),
                        "predictions": predictions,
                        "proto": proto,
                        "con": self,
                        "propogated": 0
                    }

                    # Delete candidate if it already exists.
                    if node_ip in self.factory.candidates:
                        # Max candidates reached.
                        if len(self.factory.candidates[node_ip]) >= self.max_candidates:
                            print("Candidate max candidates reached.")
                            break

                        for test_candidate in self.factory.candidates[node_ip]:
                            if test_candidate["ip_addr"] == client_ip:
                                self.factory.candidates[node_ip].remove(test_candidate)
                                print("Candidate removign test canadidate.")
                                break

                    self.factory.candidates[node_ip].append(candidate)
                    candidate_index = len(self.factory.candidates[node_ip]) - 1

                    # Update predictions.
                    self.factory.candidates[node_ip][candidate_index]["predictions"] = predictions
                    msg = "PREDICTION SET"
                    self.send_line(msg)

                    # Synchronize simultaneous node.
                    self.synchronize_simultaneous(node_ip)
                    break


            # Node wishes to respond to a simultaneous open challenge from a client.
            if re.match("^ACCEPT", line) is not None:
                # ACCEPT 192.168.0.1 4552 345 TCP 1412137849.288068
                parts = re.findall("^ACCEPT ([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+) ((?:[0-9]+\s?)+) (TCP|UDP) ([0-9]+(?:[.][0-9]+)?)$", line)
                while 1:
                    # Invalid reply.
                    if not len(parts):
                        break
                    client_ip, predictions, proto, ntp = parts[0]

                    # Invalid IP address.
                    node_ip = self.transport.getPeer().host                
                    if node_ip not in self.factory.candidates:
                        break

                    # Invalid predictions.
                    predictions = predictions.split(" ")
                    valid_ports = 1
                    for port in predictions:
                        if not self.is_valid_port(port):
                            valid_ports = 0
                    if not valid_ports:
                        break

                    # Invalid NTP.
                    ntp = ntp
                    t = time.time()
                    minute = 60 * 10
                    if int(float(ntp)) < t - minute or int(float(ntp)) > t + minute:
                        break

                    # Relay fight to client_ip.
                    # FIGHT 192.168.0.1 4552 345 34235 TCP 123123123.1
                    msg = "FIGHT %s %s %s %s" % (node_ip, " ".join(map(str, predictions)), proto, str(ntp))
                    for candidate in self.factory.candidates[node_ip]:
                        if candidate["ip_addr"] == client_ip:
                            candidate["con"].send_line(msg)

                            """
                            Signal to propogate_candidates() not to relay this candidate again. Note that this occurs after a valid accept which thus counts as acknowledging receiving the challenge.
                            """
                            candidate["propogated"] = 1
                            break

                    break

            # Remove node details.
            if re.match("^CLEAR", line) is not None:
                ip_addr = self.transport.getPeer().host
                if ip_addr in self.factory.nodes["passive"]:
                    del self.factory.nodes["passive"][ip_addr]
                if ip_addr in self.factory.nodes["simultaneous"]:
                    del self.factory.nodes["simultaneous"][ip_addr]

        except Exception as e:
            error = parse_exception(e)
            log_exception(error_log_path, error)
            print(self.log_entry("ERROR =", error))


class RendezvousFactory(Factory):
    """
    PASSIVE
    A node with the incoming port successfully reachable.
    For clients, this is done by opening the port.

    Outbound: Passive, simultaneous
    Inbound: Active, passive, simultaneous

    SIMULTANEOUS
    This is a node with a predictable NAT. TCP hole
    punching can be used to connect to these nodes.
    Note that a simultaneous node can technically
    also be a passive node and may be added to
    both categories.

    Outbound: Passive, simultaneous
    Inbound: Active, passive, simultaneous

    ACTIVE
    (For explanation only, this struct isn't used.)

    Active nodes are nodes which can only make outgoing
    connections - they cannot receive any. These nodes are
    problematic in a P2P network because a P2P network by
    definition requires nodes capable of receiving connections,
    or incoming connections to form the network. As a
    network first starts, it's likely there will be a lot
    more active nodes than anything else which will tax the
    network by taking up those limited inbound connection slots.
    Simultaneous nodes were created to try solve this problem
    and greatly make bootstrapping more automagic and user-friendly.

    Outbound: Passive, simultaneous
    Inbound:

    RELAY
    Not yet used: Relay nodes will be nodes capable of
    relaying connections on behalf of other nodes.

    BOOTSTRAP
    Not yet used: A list of other servers which can be
    used for bootstrapping.
    """

    def __init__(self):
        self.last_cleanup = time.time()
        self.candidates = {}
        self.nodes = {
            'passive': {},
            'simultaneous': {},
            'active': {},
            'relay': {},
            'bootstrap': {}
        }

        # Test data.
        """
        test_ip = "192.168.0.10"
        self.nodes["passive"][test_ip] = {
            "ip_addr": test_ip,
            "port": 194,
        }
        self.nodes["simultaneous"][test_ip] = {
            "time": time.time(),
            "ip_addr": test_ip,
            "port": 0
        }
        self.candidates[test_ip] = []
        """

    def buildProtocol(self, addr):
        return RendezvousProtocol(self)

if __name__ == "__main__":
    print("Starting rendezvous server.")
    factory = RendezvousFactory()
    reactor.listenTCP(8000, factory, interface="0.0.0.0")
    reactor.run()




#!/usr/bin/env python

"""NAT-PMP client library

Provides functions to interact with NAT-PMP gateways implementing version 0
of the NAT-PMP draft specification.

This version does not completely implement the draft standard.
* It does not provide functionality to listen for address change packets. 
* It does not have a proper request queuing system, meaning that
multiple requests may be issued in parallel, against spec recommendations.

For more information on NAT-PMP, see the NAT-PMP draft specification:

http://files.dns-sd.org/draft-cheshire-nat-pmp.txt

Requires Python 2.3 or later.
Tested on Python 2.5, 2.6 against Apple AirPort Express.

Coinbend change log:
* Changed gateway auto-detection to use netifaces which makes
this library more cross-platform.
* Added an easy to use port forwarding function with the same
interface as that seen in the UPnP module.
"""

import netifaces

__version__ = "0.2.3"
__license__ = """Copyright (c) 2008-2014, Yiming Liu, All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
* The names of the author and contributors may not be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS'
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE."""

__author__ = "Yiming Liu <http://www.yimingliu.com/>"

import struct, socket, select, time, platform
import sys, os, re

NATPMP_PORT = 5351

NATPMP_RESERVED_VAL = 0

NATPMP_PROTOCOL_UDP = 1
NATPMP_PROTOCOL_TCP = 2

NATPMP_GATEWAY_NO_VALID_GATEWAY = -10
NATPMP_GATEWAY_NO_SUPPORT = -11
NATPMP_GATEWAY_CANNOT_FIND = -12

NATPMP_RESULT_SUCCESS = 0 # Success
NATPMP_RESULT_UNSUPPORTED_VERSION = 1 # Unsupported Version
NATPMP_RESULT_NOT_AUTHORIZED = 2 # Not Authorized/Refused/NATPMP turned off
NATPMP_RESULT_NETWORK_FAILURE = 3 # Network Failure
NATPMP_RESULT_OUT_OF_RESOURCES = 4 # can not create more mappings
NATPMP_RESULT_UNSUPPORTED_OPERATION = 5 # not a supported opcode
# all remaining results are fatal errors

NATPMP_ERROR_DICT = {
                NATPMP_RESULT_SUCCESS:"No error.",
                NATPMP_RESULT_UNSUPPORTED_VERSION:"The protocol version specified is unsupported.",
                NATPMP_RESULT_NOT_AUTHORIZED:"The operation was refused.  NAT-PMP may be turned off on gateway.",
                NATPMP_RESULT_NETWORK_FAILURE:"There was a network failure.  The gateway may not have an IP address.",# Network Failure
                NATPMP_RESULT_OUT_OF_RESOURCES:"The NAT-PMP gateway is out of resources and cannot create more mappings.", # can not create more mappings
                NATPMP_RESULT_UNSUPPORTED_OPERATION:"The NAT-PMP gateway does not support this operation", # not a supported opcode
                NATPMP_GATEWAY_NO_SUPPORT:'The gateway does not support NAT-PMP',
                NATPMP_GATEWAY_NO_VALID_GATEWAY:'No valid gateway address was specified.',
                NATPMP_GATEWAY_CANNOT_FIND:'Cannot automatically determine gateway address.  Must specify manually.'
              }


class NATPMPRequest(object):
    """Represents a basic NAT-PMP request.  This currently consists of the
       1-byte fields version and opcode.
       
       Other requests are derived from NATPMPRequest.
    """
    retry_increment = 0.250 # seconds

    def __init__(self, version, opcode):
        self.version = version
        self.opcode = opcode

    def toBytes(self):
        """Converts the request object to a byte string."""
        return struct.pack('!BB', self.version, self.opcode)

class PublicAddressRequest(NATPMPRequest):
    """Represents a NAT-PMP request to the local gateway for a public address.
       As per the specification, this is a generic request with the opcode = 0.
    """
    def __init__(self, version=0):
        NATPMPRequest.__init__(self, version, 0)

class PortMapRequest(NATPMPRequest):
    """Represents a NAT-PMP request to the local gateway for a port mapping.
       As per the specification, this request extends NATPMPRequest with
       the fields private_port, public_port, and lifetime.  The first two
       are 2-byte unsigned shorts, and the last is a 4-byte unsigned integer.
    """
    def __init__(self, protocol, private_port, public_port, lifetime=3600, version=0):
        NATPMPRequest.__init__(self, version, protocol)
        self.private_port = private_port
        self.public_port = public_port
        self.lifetime = lifetime

    def toBytes(self):
        s= NATPMPRequest.toBytes(self) + struct.pack('!HHHI', NATPMP_RESERVED_VAL, self.private_port, self.public_port, self.lifetime)  
        return s

class NATPMPResponse(object):
    """Represents a generic NAT-PMP response from the local gateway.  The
       generic response has fields for version, opcode, result, and secs
       since last epoch (last boot of the NAT gateway).  As per the
       specification, the opcode is offset by 128 from the opcode of
       the original request.
    """
    def __init__(self, version, opcode, result, sec_since_epoch):
        self.version = version
        self.opcode = opcode
        self.result = result
        self.sec_since_epoch = sec_since_epoch
        
    def __str__(self):
        return "NATPMPResponse(%d, %d, %d, $d)" % (self.version, self.opcode, self.result, self.sec_since_epoch)

class PublicAddressResponse(NATPMPResponse):
    """Represents a NAT-PMP response from the local gateway to a
       public-address request.  It has one additional 4-byte field
       containing the IP returned.
       
       The member variable ip contains the Python-friendly string form, while
       ip_int contains the same in the original 4-byte unsigned int.
    """
    def __init__(self, bytes):
        if (len(bytes) > 12):
            bytes = bytes[:12]
        version, opcode, result, sec_since_epoch, self.ip_int = struct.unpack("!BBHII", bytes)
        NATPMPResponse.__init__(self, version, opcode, result, sec_since_epoch)
        self.ip = socket.inet_ntoa(bytes[8:8+4])
        #self.ip  = socket.inet_ntoa(self.ip_bytes)

    def __str__(self):
        return "PublicAddressResponse: version %d, opcode %d (%d), result %d, ssec %d, ip %s" % (self.version, self.opcode, self.result, self.sec_since_epoch, self.ip)

class PortMapResponse(NATPMPResponse):
    """Represents a NAT-PMP response from the local gateway to a
       public-address request.  The response contains the private port,
       public port, and the lifetime of the mapping in addition to typical
       NAT-PMP headers.  Note that the port mapping assigned is
       NOT NECESSARILY the port requested (see the specification
       for details).
    """
    def __init__(self, bytes):
        if len(bytes) > 16:
            bytes = bytes[:16]
        version, opcode, result, sec_since_epoch, self.private_port, self.public_port, self.lifetime = struct.unpack('!BBHIHHI', bytes)
        NATPMPResponse.__init__(self, version, opcode, result, sec_since_epoch)
    
    def __str__(self):
        return "PortMapResponse: version %d, opcode %d (%d), result %d, ssec %d, private_port %d, public port %d, lifetime %d" % (self.version, self.opcode, self.opcode, self.result, self.sec_since_epoch, self.private_port, self.public_port, self.lifetime)

class NATPMPError(Exception):
    """Generic exception state.  May be used to represent unknown errors."""
    pass

class NATPMPResultError(NATPMPError):
    """Used when a NAT gateway responds with an error-state response."""
    pass

class NATPMPNetworkError(NATPMPError):
    """Used when a network error occurred while communicating
       with the NAT gateway."""
    pass

class NATPMPUnsupportedError(NATPMPError):
    """Used when a NAT gateway does not support NAT-PMP."""
    pass


def get_gateway_addr():
    return netifaces.gateways()["default"][netifaces.AF_INET][0]

def error_str(result_code):
    """Takes a numerical error code and returns a human-readable
       error string.
    """
    result = NATPMP_ERROR_DICT.get(result_code)
    if not result:
        result = "Unknown fatal error."
    return result

def get_gateway_socket(gateway):
    """Takes a gateway address string and returns a non-blocking UDP
       socket to communicate with its NAT-PMP implementation on
       NATPMP_PORT.
       
       e.g. addr = get_gateway_socket('10.0.1.1')
    """
    if not gateway:
        raise NATPMPNetworkError(NATPMP_GATEWAY_NO_VALID_GATEWAY, error_str(NATPMP_GATEWAY_NO_VALID_GATEWAY))
    response_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    response_socket.setblocking(0)
    response_socket.connect((gateway, NATPMP_PORT))
    return response_socket

def get_public_address(gateway_ip=None, retry=9):
    """A high-level function that returns the public interface IP of
       the current host by querying the NAT-PMP gateway.  IP is
       returned as string.
       
       Takes two possible keyword arguments:
            gateway_ip - the IP to the NAT-PMP compatible gateway.
                         Defaults to using auto-detection function
                         get_gateway_addr()
            retry - the number of times to retry the request if unsuccessful.
                    Defaults to 9 as per specification.
    """
    if gateway_ip == None:
        gateway_ip = get_gateway_addr()
    addr = None
    addr_request = PublicAddressRequest()
    addr_response = send_request_with_retry(gateway_ip, addr_request, response_data_class=PublicAddressResponse, retry=retry, response_size=12)
    if addr_response.result != 0:
        #sys.stderr.write("NAT-PMP error %d: %s\n" % (addr_response.result, error_str(addr_response.result)))
        #sys.stderr.flush()
        raise NATPMPResultError(addr_response.result, error_str(addr_response.result), addr_response)
    addr = addr_response.ip
    return addr

def map_tcp_port(public_port, private_port, lifetime=3600, gateway_ip=None, retry=9, use_exception=True):
    """A high-level wrapper to map_port() that requests a mapping
       for a public TCP port on the NAT to a private TCP port on this host.
       Returns the complete response on success.
       
            public_port - the public port of the mapping requested
            private_port - the private port of the mapping requested
            lifetime - the duration of the mapping in seconds.
                       Defaults to 3600, per specification.
            gateway_ip - the IP to the NAT-PMP compatible gateway.
                        Defaults to using auto-detection function
                        get_gateway_addr()
            retry - the number of times to retry the request if unsuccessful.
                    Defaults to 9 as per specification.
            use_exception - throw an exception if an error result is
                           received from the gateway.  Defaults to True.
    """
    return map_port(NATPMP_PROTOCOL_TCP, public_port, private_port, lifetime, gateway_ip=gateway_ip, retry=retry, use_exception=use_exception)

def map_udp_port(public_port, private_port, lifetime=3600, gateway_ip=None, retry=9, use_exception=True):
    """A high-level wrapper to map_port() that requests a mapping for
       a public UDP port on the NAT to a private UDP port on this host.
       Returns the complete response on success.
       
            public_port - the public port of the mapping requested
            private_port - the private port of the mapping requested
            lifetime - the duration of the mapping in seconds.
                       Defaults to 3600, per specification.
            gateway_ip - the IP to the NAT-PMP compatible gateway.
                         Defaults to using auto-detection function
                         get_gateway_addr()
            retry - the number of times to retry the request if unsuccessful.
                    Defaults to 9 as per specification.
            use_exception - throw an exception if an error result is
                            received from the gateway.  Defaults to True.
    """
    return map_port(NATPMP_PROTOCOL_UDP, public_port, private_port, lifetime, gateway_ip=gateway_ip, retry=retry, use_exception=use_exception)

def map_port(protocol, public_port, private_port, lifetime=3600, gateway_ip=None, retry=9, use_exception=True):
    """A function to map public_port to private_port of protocol.
       Returns the complete response on success.
       
            protocol - NATPMP_PROTOCOL_UDP or NATPMP_PROTOCOL_TCP
            public_port - the public port of the mapping requested
            private_port - the private port of the mapping requested
            lifetime - the duration of the mapping in seconds.
                       Defaults to 3600, per specification.
            gateway_ip - the IP to the NAT-PMP compatible gateway.
                         Defaults to using auto-detection function
                         get_gateway_addr()
            retry - the number of times to retry the request if unsuccessful.
                    Defaults to 9 as per specification.
            use_exception - throw an exception if an error result
                            is received from the gateway.  Defaults to True.
    """
    if protocol not in [NATPMP_PROTOCOL_UDP, NATPMP_PROTOCOL_TCP]:
        raise ValueError("Must be either NATPMP_PROTOCOL_UDP or NATPMP_PROTOCOL_TCP")
    if gateway_ip == None:
        gateway_ip = get_gateway_addr()
    response = None
    port_mapping_request = PortMapRequest(protocol, private_port, public_port, lifetime)
    port_mapping_response = send_request_with_retry(gateway_ip, port_mapping_request, response_data_class=PortMapResponse, retry=retry)
    if port_mapping_response.result != 0 and use_exception:
        raise NATPMPResultError(port_mapping_response.result, error_str(port_mapping_response.result), port_mapping_response)
    return port_mapping_response


def send_request(gateway_socket, request):
    gateway_socket.sendall(request.toBytes())

def read_response(gateway_socket, timeout, responseSize=16):
    data = ""
    source_addr = ("", "")
    rlist, wlist, xlist = select.select([gateway_socket], [], [], timeout)
    if rlist:
        resp_socket = rlist[0]
        try:
            data, source_addr = resp_socket.recvfrom(responseSize)
        except Exception:
            return None, None
    return data,source_addr

def send_request_with_retry(gateway_ip, request, response_data_class=None, retry=9, response_size=16):
    gateway_socket = get_gateway_socket(gateway_ip)
    n = 1
    data = ""
    while n <= retry and not data:
        send_request(gateway_socket, request)
        data,source_addr = read_response(gateway_socket, n * request.retry_increment, responseSize=response_size)
        if data is None or source_addr[0] != gateway_ip or source_addr[1] != NATPMP_PORT:
            data = "" # discard data if source mismatch, as per specification
        n += 1
    if n >= retry and not data:
        raise NATPMPUnsupportedError(NATPMP_GATEWAY_NO_SUPPORT, error_str(NATPMP_GATEWAY_NO_SUPPORT))
    if data and response_data_class:
        data = response_data_class(data)
    return data

class NatPMP:
    def __init__(self, interface="default"):
        self.interface = interface

    def forward_port(self, proto, src_port, dest_ip, dest_port=None):
        proto = proto.upper()
        valid_protos = ["TCP", "UDP"]
        if proto not in valid_protos:
            raise Exception("Invalid protocol for forwarding.")

        valid_ports = range(1, 65535)
        if src_port not in valid_ports:
            raise Exception("Invalid port for forwarding.")

        #Source port is forwarded to same destination port number.
        if dest_port == None:
            dest_port = src_port

        if proto == "TCP":
            proto = 1
        else:
            proto = 2
        return map_port(proto, src_port, dest_port)

if __name__ == "__main__":
    #
    addr = get_public_address()
    map_resp = map_tcp_port(62001, 62001)
    print (addr)
    print (map_resp.__dict__)

    #xxxxxx = NatPMP()
    #print(xxxxxx.forward_port("TCP", 12156, "192.168.0.4"))
    #print(n.is_port_forwarded(12156, "tcp"))



"""
Handles P2P connections.
All networking functions are ultimately done through
this class.
"""

import socket
import random
import select
import netifaces
import copy
from threading import Thread
import hashlib
import re
import netifaces
import signal
import os


from .upnp import *
from .nat_pmp import NatPMP
from .lib import *
from .sock import *
from .rendezvous_client import *
from .hybrid_reply import *
from .unl import UNL
from .dht_msg import DHT

# How many times a single message can be retransmitted.
max_retransmissions = 100

# Minimum time that must pass between retransmissions.
min_retransmit_interval = 5

# A theoretical time for a message to propagate across the network.
propagation_delay = 5

# A table of message hashes for received messages.
seen_messages = {}

# How often to get new DHT messages.
dht_msg_interval = 5

# How often to re-bootstrap.
rendezvous_interval = 30 * 60

# How often to re-advertise node.
# Update bootstrapping server every 10 minutes.
advertise_interval = 10 * 60

# Time that must elapse between accepting simultaneous opens.
sim_open_interval = 2

# Bootstrapping + TCP hole punching server.
rendezvous_servers = [
    {
        "addr": "158.69.201.105",
        "port": 8000
    },
    {
        "addr": "185.86.149.128",
        "port": 8000
    }
]

# Web server running script to test port forwarding.
forwarding_servers = [
    {
        "addr": "158.69.201.105",
        "port": 80,
        "url": "/pyp2p/net.php"
    }
] 

def is_msg_old(msg, record_seen=0):
    if type(msg) == str:
        msg = msg.encode("ascii")
    response_hash = hashlib.sha256(msg).hexdigest()
    if response_hash in seen_messages:
        seen = seen_messages[response_hash]
        elapsed = int(time.time()) - seen["last"]
        if elapsed < min_retransmit_interval:
            return 1

        if seen["times"] >= max_retransmissions:
            return 1

    if record_seen:
        record_msg_hash(msg)

    return 0

def record_msg_hash(msg):
    if type(msg) == str:
        msg = msg.encode("ascii")
    response_hash = hashlib.sha256(msg).hexdigest()

    if not is_msg_old(msg):
        timestamp = int(time.time())
        if response_hash in seen_messages:
            seen = seen_messages[response_hash]
            seen["times"] += 1
            seen["last"] = timestamp
        else:
            seen_messages[response_hash] = {
                "times": 1,
                "last": timestamp
            }

        return 1
    else:
        return 0

def clear_seen_messages():
    global seen_messages
    seen_messages = {}

class Net():
    def __init__(self, net_type="p2p", nat_type="unknown", node_type="unknown", max_outbound=10, max_inbound=10, passive_bind="0.0.0.0", passive_port=50500, interface="default", dht_node=None, error_log_path="error.log", debug=0):
        # List of outbound connections (from us, to another node.)
        self.outbound = []

        # List of inbound connections (to us, from another node.)
        self.inbound = []

        # Socket to receive inbound connections on.
        self.passive = None

        # Type of node: simultaneous, active, passive.
        self.node_type = node_type

        # NAT type: preserving, delta, reuse, random.
        self.nat_type = nat_type

        # Address to listen() on for inbound cons.
        self.passive_bind = passive_bind

        # Port to listen() on for inbound cons.
        self.passive_port = int(passive_port)

        # How many connections can we accept from other nodes?
        self.max_outbound = int(max_outbound)

        # How many connections can we make to other nodes?
        self.max_inbound = int(max_inbound)

        # List of servers to do port forwarding checks.
        self.forwarding_servers = forwarding_servers

        # Unix timestamp of last bootstrap.
        self.last_bootstrap = None

        # Unix timestamp of last DHT direct message.
        self.last_dht_msg = None

        # Unix timestamp of last advertise.
        self.last_advertise = None

        # What interface to make outbound connections from?
        self.interface = interface

        # Skip advertise if we have at least this many inbound connections.
        self.min_connected = 3

        # Unix timestamp of last simultaneous open challenge.
        self.last_passive_sim_open = 0

        # Does this Net instance need to bootstrap?
        self.enable_bootstrap = 1

        # Does this Net instance need to advertise?
        self.enable_advertise = 1

        # Should we try open ports?
        self.enable_forwarding = 1

        # Is simultaneous open enabled?
        self.enable_simultaneous = 1

        # Does this Net instance reject duplicate messages
        # (same hash as previous messages)?
        self.enable_duplicates = 1

        # Where should I store errors?
        self.error_log_path = error_log_path

        # Indicates port forwarding state.
        self.forwarding_type = "manual"

        # Debug mode shows debug messages.
        self.debug = debug

        # Network: p2p or direct.
        self.net_type = net_type

        # Rendezvous / boostrapping client.
        self.rendezvous = RendezvousClient(self.nat_type, rendezvous_servers=rendezvous_servers, interface=self.interface)

        # DHT node for receiving direct messages from other nodes.
        self.dht_node = dht_node

        # Node type details only known after network is start()ed.
        self.unl = None

        # List of connections that still need to respond to
        # our reverse query.
        self.pending_reverse_queries = []

        # Time frame for connection to respond to reverse query.
        self.reverse_query_expiry = 5

        # Enable more than one connection to the same IP.
        self.enable_duplicate_ip_cons = 0

        # Net instances hide their con details to prioritise direct cons.
        if self.net_type == "direct":
            self.disable_bootstrap()    
            self.disable_advertise()
            self.enable_duplicate_ip_cons = 1

        # Set to 1 when self.start() has been called.
        self.is_net_started = 0

    # Outputs debug statements if debug mode is enabled.
    def debug_print(self, msg):
        if self.debug:
            print("> " + str(msg))

    def disable_duplicates(self):
        self.enable_duplicates = 0

    def disable_bootstrap(self):
        self.enable_bootstrap = 0

    def disable_advertise(self):
        self.enable_advertise = 0
    
    def disable_simultaneous(self):
        self.enable_simultaneous = 0

    def disable_forwarding(self):
        self.enable_forwarding = 0

    def get_connection_no(self):
        return (len(self.outbound) + len(self.inbound))

    # Used to reject duplicate connections.
    def validate_node(self, node_ip, node_port=None, same_nodes=1):
        self.debug_print("Validating: " + node_ip)

        # Is this a valid IP?
        if not is_ip_valid(node_ip) or node_ip == "0.0.0.0":
            self.debug_print("Invalid node ip in validate node")
            return 0

        # Is this a valid port?
        if node_port != 0 and node_port != None:
            if not is_valid_port(node_port):
                self.debug_print("Invalid node port in validate port")
                return 0

        """
        Don't accept connections from self to passive server
        or connections to already connected nodes.
        """
        if not self.enable_duplicate_ip_cons:
            # Don't connect to ourself.
            if node_ip == "127.0.0.1" or node_ip == get_lan_ip(self.interface) or node_ip == get_wan_ip():
                self.debug_print("Cannot connect to ourself.")
                return 0

            # No, really: don't connect to ourself.
            if node_ip == self.passive_bind and node_port == self.passive_port:
                self.debug_print("Error connecting to same listen server.")
                return 0

            # Don't connect to same nodes.
            if same_nodes:
                for node in self.outbound + self.inbound:
                    try:
                        addr, port = node["con"].s.getpeername()
                        if node_ip == addr:
                            self.debug_print("Already connected to this node.")
                            return 0
                    except Exception as e:
                        print(e)
                        return 0
        
        return 1

    # Make an outbound con to a passive or simultaneous node.
    def add_node(self, node_ip, node_port, node_type, timeout=5):
        # Correct type for port.
        node_port = int(node_port)

        # Debug info.
        msg = "Attempting to connect to %s:%s:%s" % (node_ip, str(node_port), node_type)
        self.debug_print(msg)

        # Already connected to them.
        if not self.enable_duplicate_ip_cons:
            for node in self.outbound + self.inbound:
                if node_ip == node["ip"]:
                    self.debug_print("Already connected.")
                    return node["con"]

        # Avoid connecting to ourself.
        if not self.validate_node(node_ip, node_port):
            self.debug_print("Validate node failed.")
            return None
        
        # Make a simultaneous open connection.
        con = None
        if node_type == "simultaneous" and self.enable_simultaneous:
            # Check they've started net first
            # If they haven't we won't know the NAT details / node type.
            if not self.is_net_started:
                raise Exception("Make sure to start net before you add node.")

            if self.nat_type in self.rendezvous.predictable_nats:
                # Attempt to make active simultaneous connection.
                old_timeout = self.rendezvous.timeout
                try:
                    self.rendezvous.timeout = timeout
                    con = self.rendezvous.simultaneous_challenge(node_ip, node_port, "TCP")
                except Exception as e:
                    error = parse_exception(e)
                    log_exception(self.error_log_path, error)
                    return None
                self.rendezvous.timeout = old_timeout

                # Record node details and return con.
                self.rendezvous.simultaneous_cons = []
                if con != None:
                    node = {
                        "con": con,
                        "type": "simultaneous",
                        "ip": node_ip,
                        "port": 0
                    }
                    self.outbound.append(node)
                    self.debug_print("SUCCESS")
                else:
                    self.debug_print("FAILURE")

        # Passive outbound -- easiest to connect to.
        if node_type == "passive":
            try:
                # Try connect to passive server.
                con = Sock(node_ip, node_port, blocking=0, timeout=timeout, interface=self.interface)
                node = {
                    "con": con,
                    "type": "passive",
                    "ip": node_ip,
                    "port": node_port
                }
                self.outbound.append(node)
                self.debug_print("SUCCESS")
            except Exception as e:
                self.debug_print("FAILURE")
                error = parse_exception(e)
                self.debug_print(error)
                log_exception(self.error_log_path, error)
                return None

        # Filter(): Remove undesirable messages from replies.
        # Save message: 0 = no, 1 = yes.
        def filter_msg_check_builder():
            def filter_msg_check(msg):
                # Allow duplicate replies?
                record_seen = not self.enable_duplicates

                # Is this a reply to a reverse query?
                for reverse_query in self.pending_reverse_queries:
                    pattern  = "^REVERSE_ORIGIN:" + reverse_query["unl"]
                    pattern += "$"
                    if re.match(pattern, msg) != None:
                        self.pending_reverse_queries.remove(reverse_query)
                        return 0

                # Check if message is old.
                return not is_msg_old(msg, record_seen)

            return filter_msg_check

        # Patch sock object to reject duplicate replies.
        # (if enabled)
        con.reply_filter = filter_msg_check_builder()

        # Return new connection.
        return con

    def bootstrap(self):
        """
        When the software is first started, it needs to retrieve
        a list of nodes to connect to the network to. This function
        asks the server for N nodes which consists of at least N
        passive nodes and N simultaneous nodes. The simultaneous
        nodes are prioritized if the node_type for the machine
        running this software is simultaneous, with passive nodes
        being used as a fallback. Otherwise, the node exclusively
        uses passive nodes to bootstrap.

        This algorithm is designed to preserve passive node's
        inbound connection slots.
        """
        # Disable bootstrap.
        if not self.enable_bootstrap:
            return

        # Avoid raping the rendezvous server.
        t = time.time()
        if self.last_bootstrap != None:
            if t - self.last_bootstrap <= rendezvous_interval:
                return
        self.last_bootstrap = t
        self.debug_print("Searching for nodes to connect to.")

        try:
            connection_slots = self.max_outbound - (len(self.outbound))
            passive_outbound = []
            simultaneous_outbound = []
            if connection_slots > 0:
                # Connect to rendezvous server.
                rendezvous_con = Sock(self.rendezvous.rendezvous_servers[0]["addr"], self.rendezvous.rendezvous_servers[0]["port"], blocking=1, interface=self.interface, timeout=3)

                # Retrieve random nodes to bootstrap with.
                rendezvous_con.send_line("BOOTSTRAP " + str(self.max_outbound * 2))
                choices = rendezvous_con.recv_line(timeout=2)
                if choices == "NODES EMPTY":
                    rendezvous_con.close()
                    self.debug_print("Node list is empty.")
                    return
                else:
                    self.debug_print("Found node list.")

                # Parse node list.
                choices = re.findall("(?:(p|s)[:]([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+)[:]([0-9]+))+\s?", choices)
                rendezvous_con.s.close()

                # Attempt to make active simultaneous connections.
                passive_nodes = []
                for node in choices:
                    # Out of connection slots.
                    if not connection_slots:
                        break

                    """
                    Note: This code has been disabled because bootstrapping with simultaneous nodes was never a good idea (plus this code is blocking.) Instead, TCP hole punching is reserved as a last resort.

                    # Attempt TCP hole punch (blocking.)
                    node_type, node_ip, node_port = node
                    if node_type == "s":
                        con = self.add_node(node_ip, node_port, "simultaneous")
                        if con != None:
                            connection_slots -= 1
                            self.debug_print("Con successful.")
                        else:
                            self.debug_print("Con failed.")
                    """

                    # Add to list of passive nodes.
                    if node_type == "p":
                        passive_nodes.append(node)

                # Use passive to make up the remaining cons.
                i = 0
                while i < len(passive_nodes) and connection_slots > 0:
                    node_type, node_ip, node_port = passive_nodes[i]
                    con = self.add_node(node_ip, node_port, "passive")
                    if con != None:
                        connection_slots -= 1
                        self.debug_print("Con successful.")
                    else:
                        self.debug_print("Con failed.")

                    i += 1

        except Exception as e:
            error = parse_exception(e)
            log_exception(self.error_log_path, error)

    def advertise(self):
        """
        This function tells the rendezvous server that our node is ready to accept connections from other nodes on the P2P network that run the bootstrap function. It's only used when net_type == p2p
        """

        # Advertise is disabled.
        if not self.enable_advertise:
            return

        # Direct net is reserved for direct connections only.
        if self.net_type == "direct":
            return

        # Net isn't started!.
        if not self.is_net_started:
            raise Exception("Please call start() before you call advertise()")

        # Avoid raping the rendezvous server with excessive requests.
        t = time.time()
        if self.last_advertise != None:
            if t - self.last_advertise <= advertise_interval:
                return

            if len(self.inbound) >= self.min_connected:
                return

        self.last_advertise = t

        # Tell rendezvous server to list us.
        try:
            # We're a passive node.
            if self.node_type == "passive" and self.passive_port != None and self.enable_advertise:
                self.rendezvous.passive_listen(self.passive_port, self.max_inbound)

            """
            Simultaneous open is only used as a fail-safe for connections to nodes on the direct_net and only direct_net can list itself as simultaneous so its safe to leave this enabled.
            """
            if self.node_type == "simultaneous":
                self.rendezvous.simultaneous_listen()
        except Exception as e:
            error = parse_exception(e)
            log_exception(self.error_log_path, error)

    def determine_node(self):
        """
        Determines the type of node based on a combination of forwarding reachability and NAT type.
        """

        # Manually set node_type as simultaneous.
        if self.node_type == "simultaneous":
            if self.nat_type != "unknown":
                return "simultaneous"

        # Get IP of binding interface.
        unspecific_bind = ["0.0.0.0", "127.0.0.1", "localhost"]
        if self.passive_bind in unspecific_bind:
            lan_ip = get_lan_ip(self.interface)
        else:
            lan_ip = self.passive_bind

        # Passive node checks.
        if lan_ip != None and self.passive_port != None and self.enable_forwarding:
            self.debug_print("Checking if port is forwarded.")

            # Check port isn't already forwarded.
            if is_port_forwarded(lan_ip, self.passive_port, "TCP", self.forwarding_servers):
                self.debug_print("Port already forwarded. Skipping NAT traversal.")

                self.forwarding_type = "forwarded"
                return "passive"
            else:
                self.debug_print("Port is not already forwarded.")

            # Most routers.
            try:
                self.debug_print("Trying UPnP")

                UPnP(self.interface).forward_port("TCP", self.passive_port, lan_ip)

                if is_port_forwarded(lan_ip, self.passive_port, "TCP", self.forwarding_servers):
                    self.forwarding_type = "UPnP"
                    self.debug_print("Forwarded port with UPnP.")
                else:
                    self.debug_print("UPnP failed to forward port.")

            except Exception as e:
                # Log exception.
                error = parse_exception(e)
                log_exception(self.error_log_path, error)
                self.debug_print("UPnP failed to forward port.")

                # Apple devices.
                try:
                    self.debug_print("Trying NATPMP.")
                    NatPMP(self.interface).forward_port("TCP", self.passive_port, lan_ip)
                    if is_port_forwarded(lan_ip, self.passive_port, "TCP", self.forwarding_servers):
                        self.forwarding_type = "NATPMP"
                        self.debug_print("Port forwarded with NATPMP.")
                    else:
                        self.debug_print("Failed to forward port with NATPMP.")
                        self.debug_print("Falling back on TCP hole punching or proxying.")
                except Exception as e:
                    # Log exception
                    error = parse_exception(e)
                    log_exception(self.error_log_path, error)
                    self.debug_print("Failed to forward port with NATPMP.")

            # Check it worked.
            if self.forwarding_type != "manual":
                return "passive"

        # Fail-safe node types.
        if self.nat_type != "unknown":
            return "simultaneous"
        else:
            return "active"

    # Receive inbound connections.
    def start_passive_server(self):
        self.passive = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.passive.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.passive.bind((self.passive_bind, self.passive_port))
        self.passive.listen(self.max_inbound)

    def start(self):
        """
        This function determines node and NAT type, saves connectivity details, and starts any needed servers to be a part of the network. This is usually the first function called after initialising the Net class.
        """

        self.debug_print("Starting networking.")
        self.debug_print("Make sure to iterate over replies if you need connection alive management!")

        # Register a cnt + c handler
        signal.signal(signal.SIGINT, self.stop)

        # Check this netifaces version supports gateway function.
        # Older versions don't.
        try:
            netifaces.gateways()
        except AttributeError:
            raise Exception("You must be using at least version 0.10.0 of netifaces.")

        # Save WAN IP.
        self.debug_print("WAN IP = " + str(get_wan_ip()))

        # Check rendezvous server is up.
        try:
            rendezvous_con = Sock(self.rendezvous.rendezvous_servers[0]["addr"], self.rendezvous.rendezvous_servers[0]["port"], blocking=1, interface=self.interface, timeout=3)
            rendezvous_con.close()
        except:
            raise Exception("Unable to connect to rendezvous server.")

        # Determine NAT type.
        if self.nat_type == "unknown":
            self.debug_print("Determining NAT type.")
            nat_type = self.rendezvous.determine_nat()
            if nat_type != None and nat_type != "unknown":
                self.nat_type = nat_type
                self.rendezvous.nat_type = nat_type
                self.debug_print("NAT type = " + nat_type)
            else:
                self.debug_print("Unable to determine NAT type.")

        # Check NAT type if node is simultaneous
        # is manually specified.
        if self.node_type == "simultaneous":
            if self.nat_type not in self.rendezvous.predictable_nats:
                self.debug_print("Manual setting of simultanous specified but ignored since NAT does not support it.")
                self.node_type = "unknown"
        else:
            # Determine node type.
            self.debug_print("Determining node type.")
            self.start_passive_server()

            # No checks for manually specifying passive
            # (there probably should be.)
            if (self.node_type != "passive" and self.debug != 1) or self.node_type == "unknown":
                self.node_type = self.determine_node()
                if self.net_type == "p2p":
                    """
                    TCP hole punching is reserved specifically for direct networks (a net object reserved for receiving direct connections -- p2p is for connecting to the main network. The reason for this is you can't do multiple TCP hole punches at the same time so its reserved for direct network where it's most needed.
                    """
                    if self.node_type == "simultaneous":
                        self.node_type = "active"
                        self.disable_simultaneous()

        self.debug_print("Node type = " + self.node_type)

        # Close stray cons from determine_node() tests.
        self.close_cons()

        # Set net started status.
        self.is_net_started = 1

        # Initialise our UNL details.
        self.unl = UNL(self, self.dht_node)

    def stop(self, signum=None, frame=None):
        self.debug_print("Stopping networking.")

        if self.passive != None:
            self.passive.shutdown(1)
            self.passive.close()
            self.passive = None

        if self.last_advertise != None:
            self.rendezvous.leave_fight()

        for con in self:
            con.close()

        if signum != None:
            raise Exception("Process was interrupted.")

    # Return a connection by its IP.
    def con_by_ip(self, ip):
        for node in self.outbound + self.inbound:
            if node["ip"] == ip:
                return node["con"]

        return None
    
    # Send a message to all currently established connections.
    def broadcast(self, msg, source_con=None):
        for node in self.outbound + self.inbound:
            if node["con"] != source_con:
                node["con"].send_line(msg)

    def close_cons(self):
        # Close all connections.
        for node in self.inbound + self.outbound:
            node["con"].close()

        # Flush client queue for passive server.
        if self.node_type == "passive" and self.passive != None:
            self.passive.close()
            self.start_passive_server()

        # Start from scratch.
        self.inbound = []
        self.outbound = []

    def synchronize(self):
        # Clean up dead connections.
        for node_list_name in ["self.inbound", "self.outbound"]:
            node_list = eval(node_list_name)[:]
            for node in node_list:
                if not node["con"].connected:
                    self.debug_print("Removing disconnected: " + str(node))
                    eval(node_list_name).remove(node)

        # Timeout connections that haven't responded to reverse query.
        old_reverse_queries = []
        for reverse_query in self.pending_reverse_queries:
            duration = time.time() - reverse_query["timestamp"]
            if duration >= self.reverse_query_expiry:
                reverse_query["con"].close()
                old_reverse_queries.append(reverse_query)

        # Remove old reverse queries.
        for reverse_query in old_reverse_queries:
            self.pending_reverse_queries.remove(reverse_query)

        # Check for reverse connect requests.
        if self.dht_node != None and self.net_type == "direct":
            # Don't do this every synch cycle.
            t = time.time()
            skip_dht_check = 0
            if self.last_dht_msg != None:
                if t - self.last_dht_msg > dht_msg_interval:
                    skip_dht_check = 1

            if not skip_dht_check and self.dht_node.has_messages():
                dht_messages = []
                for msg in self.dht_node.get_messages():
                    # Found reverse connect request.
                    if re.match("^REVERSE_CONNECT:[a-zA-Z0-9+/-=_\s]+$", msg) != None:
                        call, their_unl = msg.split(":")

                        # Ask if the source sent it.
                        def success_builder():
                            def success(con):
                                # Did you send this?
                                con.send_line("REVERSE_QUERY:" + self.unl.value)

                                # Record pending query state.
                                query = {
                                    "unl": their_unl,
                                    "con": con,
                                    "timestamp": time.time()
                                }
                                self.pending_reverse_queries.append(query)

                            return success

                        self.unl.connect(their_unl, {"success": success_builder()}, reverse_connect=1)
                    else:
                        dht_messages.append(msg)

                # Put messages back.
                for msg in dht_messages:
                    self.debug_print(msg)
                    self.dht_node.messages_received.put_nowait(msg)

            self.last_dht_msg = t

        # Accept inbound connections.
        if len(self.inbound) < self.max_inbound:
            # Accept new passive inbound connections.
            if self.passive != None:
                r, w, e = select.select([self.passive], [], [], 0)
                for s in r:
                    if s == self.passive:
                        # Accept a new con from the listen queue.
                        client, address = self.passive.accept()
                        con = Sock()
                        con.set_sock(client)
                        node_ip, node_port = con.s.getpeername()

                        # Reject duplicate connections.
                        if self.validate_node(node_ip, node_port):
                            node = {
                                "type": "accept",
                                "con": con,
                                "ip": con.s.getpeername()[0],
                                "port": con.s.getpeername()[1],
                            }
                            self.inbound.append(node)
                            self.debug_print("Accepted new passive connection: " + str(node))
                        else:
                            con.close()

            # Accept new passive simultaneous connections.
            if self.node_type == "simultaneous":
                """
                This is basically the code that passive simultaneous
                nodes periodically call to parse any responses from the
                Rendezvous Server which should hopefully be new
                requests to initiate hole punching from active
                simultaneous nodes.

                If a challenge comes in, the passive simultaneous
                node accepts the challenge by giving details to the
                server for the challenging node (active simultaneous)
                to complete the simultaneous open.
                """

                # try:
                t = time.time()
                if self.rendezvous.server_con != None:
                    for reply in self.rendezvous.server_con:
                        # Reconnect.
                        if re.match("^RECONNECT$", reply) != None:
                            if self.enable_advertise:
                                self.rendezvous.simultaneous_listen()
                            continue

                        # Find any challenges.
                        # CHALLENGE 192.168.0.1 50184 50185 50186 50187 TCP
                        parts = re.findall("^CHALLENGE ([0-9]+[.][0-9]+[.][0-9]+[.][0-9]+) ((?:[0-9]+\s?)+) (TCP|UDP)$", reply)
                        if not len(parts):
                            continue
                        candidate_ip, candidate_predictions, candidate_proto = parts[0]

                        # Already connected.
                        if not self.validate_node(candidate_ip):
                            continue

                        # Last meeting was too recent.
                        if t - self.last_passive_sim_open < sim_open_interval:
                            continue

                        # Accept challenge.
                        our_ntp = get_ntp()
                        if our_ntp == None:
                            continue
                        msg = "ACCEPT %s %s TCP %s" % (candidate_ip, self.rendezvous.predictions, str(our_ntp))
                        ret = self.rendezvous.server_con.send_line(msg)
                        if not ret:
                            continue

                        """
                        Adding threading here doesn't work because Python's
                        fake threads and the act of starting a thread ruins
                        the timing between code synchronisation - especially
                        code running on the same host or in a LAN. Will
                        compensate by reducing the NTP delay to have the
                        meetings occur faster and setting a limit for meetings
                        to occur within the same period. 
                        """
                        # Walk to fight and return holes made.
                        self.last_passive_sim_open = t
                        con = self.rendezvous.attend_fight(self.rendezvous.mappings, candidate_ip, candidate_predictions, our_ntp)
                        if con != None:
                            node = {
                                "type": "simultaneous",
                                "con": con,
                                "ip": con.s.getpeername()[0],
                                "port": con.s.getpeername()[1],
                            }
                            self.inbound.append(node)

                        # Create new predictions ready to accept next client.
                        self.rendezvous.simultaneous_cons = []
                        if self.enable_advertise:
                            self.rendezvous.simultaneous_listen()

        # QUIT - remove us from bootstrapping server.
        if len(self.inbound) == self.max_inbound:
            try:
                # Remove advertise.
                self.rendezvous.leave_fight()
            except:
                pass

        # Bootstrap again if needed.
        self.bootstrap()

        # Relist node again if noded.
        self.advertise()
                    
    """
    These functions here make the class behave like a list. The
    list is a collection of connections (inbound) + (outbound.)
    Every iteration also has the bonus of reaping dead connections,
    making new ones (if needed), and accepting connections
    """
    def __len__(self):
        self.synchronize()
        return len(self.inbound) + len(self.outbound)

    def __iter__(self):
        # Process connections.
        self.synchronize()

        # Copy all connections to single buffer.
        cons = []
        for node in self.inbound:
            cons.append(node["con"])
        for node in self.outbound:
            cons.append(node["con"])

        # Return all cons.
        return iter(cons)

if __name__ == "__main__":

    """
    net = Net(debug=1)
    net.disable_bootstrap()
    net.disable_advertise()
    net.disable_forwarding()
    net.start()
    print(net.unl.value)
    print(net.unl.deconstruct(net.unl.value))

    while 1:
        time.sleep(0.5)
    """

    # Test simultaneous open.
    p2p_net = Net(debug=1, nat_type="preserving", node_type="simultaneous")
    p2p_net.start()
    p2p_net.disable_advertise()
    p2p_net.disable_bootstrap()
    # p2p_net.add_node("192.187.97.131", 0, "simultaneous") # Behind NAT
    # print(p2p_net.add_node("162.218.239.6", 0, "simultaneous")) # Not behind NAT

    def success_notify(con):
        print("SUCCESS THREADING.")


    #Test UNL
    events = {
        "success": success_notify
    }
    #print(p2p_net.unl.deconstruct("AQAAAAAAAAAAAAAAAAAAAAAAAAAAc2dtRMUG79qiBu/aos6tMVYAAAAAWMYQkz0OjrI="))
    #exit()


    #p2p_net.unl.connect("AQAAAAAAAAAAAAAAAAAAAAAAAAAAc2dtRMUG79qiBu/aos6tMVYAAAAAWMYQkz0OjrI=", events)

    while 1:
        time.sleep(0.5)

    exit()



    #P2P network example.
    p2p_net = Net(debug=1)
    p2p_net.start()
    p2p_net.bootstrap()
    p2p_net.advertise()

    #Event loop.
    while 1:
        for con in p2p_net:
            for reply in con:
                print(reply)

            """
            Excluses con from broadcast since we got this message from them
            """
            p2p_net.broadcast("Something.", con)

        time.sleep(0.5)

    #Direct network example.
    dht_node = DHT()
    direct_net = Net(dht_node=dht_node, debug=1)
    direct_net.start()

    #Connect to some UNL.
    def success(con):
        con.send_line("Thanks.")

    #Note: this isn't a valid UNL.
    #To get your UNL do: direct_net.unl.value.
    direct_net.unl.connect("Some guys UNL...", {"success": success})




"""
Universal Node Locator (UNL.) Allows nodes to direct connect and helps to debug issues in doing so.
"""

from .lib import *
import time
import struct
import random
import binascii
import base64
import binascii
from threading import Thread, Lock

def is_valid_unl(value):
    unl = UNL(value=value)
    ret = unl.deconstruct()
    if ret == None:
        return 0
    else:
        return 1

class UNL():
    def __init__(self, net=None, dht_node=None, value=None):
        self.version = 1
        self.net = net
        self.nat_type_lookup = {
            "m": "random",
            "g": "preserving",
            "e": "reuse"
        }

        self.forwarding_type_lookup = {
            "f": "forwarded",
            "m": "manual",
            "U": "UPnP",
            "N": "NATPMP"
        }

        self.node_type_lookup = {
            "p": "passive",
            "a": "active",
            "s": "simultaneous"
        }

        self.dht_node = dht_node
        if value != None:
            self.value = value
        else:
            self.value = self.construct()

        # Table of in progress UNLs.
        self.pending_unls = []

        # Sim opens are queued to occur sequentially.
        self.pending_sim_open = []

        # Simple mutex.
        self.mutex = Lock()

    def __eq__(self, other):
        # Compare UNLs. Used to check if a UNL is us.
        our_unl = self.deconstruct(self.value)
        their_unl = self.deconstruct(other.value)

        # Different WAN IPs.
        if our_unl["wan_ip"] != their_unl["wan_ip"]:
            return False

        # Different LAN IPs.
        if our_unl["lan_ip"] != their_unl["lan_ip"]:
            return False

        # Different passive ports.
        if our_unl["listen_port"] != their_unl["listen_port"]:
            return False

        # They are the same.
        return True

    # Operator !=
    def __ne__(self, other):
        return not self == other

    def connect_handler(self, their_unl, events, force_master):
        # Figure out who should make the connection.
        our_unl = self.value.encode("ascii")
        their_unl = their_unl.encode("ascii")
        int_our_unl = int(binascii.hexlify(our_unl), 16)
        int_their_unl = int(binascii.hexlify(their_unl), 16)
        if int_our_unl > int_their_unl:
            master = 1
        else:
            master = 0

        """
        Master defines who connects if either side can. It's used to eliminate having multiple connections with the same host.
        """
        if force_master:
            master = 1

        # Deconstruct binary UNLs into dicts.
        our_unl = self.deconstruct(our_unl)
        their_unl = self.deconstruct(their_unl)

        if our_unl == None:
            raise Exception("Unable to deconstruct our UNL.")
        if their_unl == None:
            raise Exception("Unable to deconstruct their UNL.")

        # This means the nodes are behind the same router.
        if our_unl["wan_ip"] == their_unl["wan_ip"]:
            # Connect to LAN IP.
            our_unl["wan_ip"] = our_unl["lan_ip"]
            their_unl["wan_ip"] = their_unl["lan_ip"]

            # Already behind NAT so no forwarding needed.
            our_unl["node_type"] = "passive"
            their_unl["node_type"] = "passive"

        # Aquire mutex.
        self.mutex.acquire()

        # Wait for other UNLs to finish.
        while their_unl in self.pending_unls:
            # This is an undifferentiated duplicate.
            if events == None:
                return

            time.sleep(1)

        # Wait for any other hole punches to finish.
        if their_unl["node_type"] == "simultaneous" and our_unl["node_type"] != "passive":
            self.pending_sim_open.append(their_unl["value"])
            print("Waiting for queued sim open")
            while len(self.pending_sim_open):
                if self.pending_sim_open[0] == their_unl["value"]:
                    break

                time.sleep(1)
            print("Ok, it's our turn now.")

        # Set pending UNL.
        self.pending_unls.append(their_unl)

        # Release mutex.
        self.mutex.release()

        # Are we asking them to connect to us?
        reverse_query = 0

        # Are they already connected?
        con = self.net.con_by_ip(their_unl["wan_ip"])
        if con == None:
            # Valid node types.
            for node_type in ["passive", "simultaneous"]:
                # Matches for this node type.
                nodes = []
                if our_unl["node_type"] == node_type:
                    nodes.append(our_unl)

                if their_unl["node_type"] == node_type:
                    nodes.append(their_unl)

                # Try the next node type.
                if len(nodes):
                    # We only want one connection.
                    if len(nodes) == 2:
                        if not master:
                            # They will connect to us.
                            nodes.remove(their_unl)
                        else:
                            # We will connect to them.
                            nodes.remove(our_unl)

                    # Don't connect to ourself.
                    node = nodes[0]
                    if node == their_unl:
                        con = self.net.add_node(their_unl["wan_ip"], their_unl["listen_port"], their_unl["node_type"], timeout=60)
                        if con != None:
                            break
                    else:
                        # Tell them to connect to us.
                        if self.dht_node != None and force_master:
                            con_request = "REVERSE_CONNECT:%s" % (self.value)
                            node_id = their_unl["node_id"]
                            if int(binascii.hexlify(node_id), 16):
                                reverse_query = 1
                                self.dht_node.send_direct_message(node_id, con_request)

                        # They will connect to us.
                        found_con = 0
                        for i in range(0, 60):
                            con = self.net.con_by_ip(their_unl["wan_ip"])
                            if con != None:
                                found_con = 1
                                break

                            time.sleep(1)

                        if found_con:
                            break

        # Undo pending connect state.
        self.pending_unls.remove(their_unl)

        # Undo pending sim open.
        if len(self.pending_sim_open):
            if self.pending_sim_open[0] == their_unl["value"]:
                self.pending_sim_open = self.pending_sim_open[1:]

        # Only execute events if this function was called manually.
        if events != None:
            # Success.
            if con != None:
                # Confirm we sent the reverse connect request.
                if reverse_query:
                    msg = con.recv_line(timeout=2)
                    if "REVERSE_QUERY:" in msg:
                        reverse_accept = "REVERSE_ORIGIN:" + self.value
                        con.send_line(reverse_accept)
                    else:
                        con.close()
                        del events["success"]
                        con = None

                if "success" in events:
                    events["success"](con)

            # Failure.
            if con == None:
                if "failure" in events:
                    events["failure"](con)

    def connect(self, their_unl, events, force_master=1):
        """
        A new thread is spawned because many of the connection techniques rely on sleep to determine connection outcome or to synchronise hole punching techniques. If the sleep is in its own thread it won't block main execution.
        """
        parms = (their_unl, events, force_master)
        t = Thread(target=self.connect_handler, args=parms)
        t.start()

    def deconstruct(self, unl=None):
        if unl == None:
            unl = self.value

        try:
            if sys.version_info >= (3,0,0):
                # for Python 3
                if isinstance(unl, bytes):
                    unl = unl.decode('ascii')  # or  s = str(s)[2:-1]
            else:
                # for Python 2
                if isinstance(unl, str):
                    unl = unicode(unl)

            # Separate checksum.
            value = unl
            unl = base64.b64decode(unl)
            checksum_size = 4
            checksum = unl[-checksum_size:]
            unl = unl[:-checksum_size]

            # Check checksum.
            expected_checksum = hashlib.sha256(hashlib.sha256(unl).digest()).digest()
            expected_checksum = expected_checksum[0:4]
            if checksum != expected_checksum:
                raise Exception("Invalid checksum -- UNL is probably corrupt.")

            # Separate the other fields.
            version, node_id, node_type, nat_type, forwarding_type, passive_port, wan_ip, lan_ip, timestamp, nonce = struct.unpack("<B20sBBBHIIQI", unl)
            node_type = chr(node_type)
            node_type = self.node_type_lookup[node_type]
            nat_type = chr(nat_type)
            nat_type = self.nat_type_lookup[nat_type]
            forwarding_type = chr(forwarding_type)
            forwarding_type = self.forwarding_type_lookup[forwarding_type]
            wan_ip = int2ip(wan_ip)
            lan_ip = int2ip(lan_ip)

            # Return meaningful fields.
            ret = {
                "value": value,
                "version": version,
                "node_id": node_id,
                "node_type": node_type,
                "nat_type": nat_type,
                "forwarding_type": forwarding_type,
                "listen_port": passive_port,
                "wan_ip": wan_ip,
                "lan_ip": lan_ip,
                "timestamp": timestamp,
                "nonce": nonce
            }

            return ret
        except Exception as e:
            print(e)
            return None

    def construct(self, details={}):
        # Sanity check.
        if self.net == None:
            raise Exception("Missing Net object for UNL.construct")

        # Translate bind address.
        wan_ip = get_wan_ip()
        if "wan_ip" in details:
            wan_ip = details["wan_ip"]

        # Lan IP.
        unspecific_bind = ["0.0.0.0", "127.0.0.1", "localhost"]
        if self.net.passive_bind in unspecific_bind:
            lan_ip = get_lan_ip(self.net.interface)
        else:
            lan_ip = self.net.passive_bind
        if "lan_ip" in details:
            lan_ip = details["lan_ip"]

        # Node id.
        if self.dht_node == None:
            node_id = b"\0"
        else:
            node_id = self.dht_node.get_id()
        if "node_id" in details:
            node_id = details["node_id"]

        # Version.
        version = self.version
        if "version" in details:
            version = details["version"]

        # Node type.
        node_type = self.net.node_type
        if "node_type" in details:
            node_type = details["node_type"]

        # Nat type.
        nat_type = self.net.nat_type
        if "nat_type" in details:
            nat_type = details["nat_type"]

        # Forwarding type.
        forwarding_type = self.net.forwarding_type
        if "forwarding_type" in details:
            forwarding_type = details["forwarding_type"]

        # Listen port.
        listen_port = self.net.passive_port
        if "listen_port" in details:
            listen_port = details["listen_port"]

        # Timestamp.
        timestamp = time.time()
        if "timestamp" in details:
            timestamp = details["timestamp"]

        # Nonce.
        nonce = random.randrange(0, 2 ** (4 * 8))
        if "nonce" in details:
            nonce = details["nonce"]

        # Generate UNL.
        unl = struct.pack("<B20sBBBHIIQI", \
            version,
            node_id,
            ord(node_type[0]),
            ord(nat_type[-1]),
            ord(forwarding_type[0]),
            listen_port,
            ip2int(wan_ip),
            ip2int(lan_ip),
            int(timestamp),
            nonce
        )

        # Build checksum and make base64.
        checksum = hashlib.sha256(hashlib.sha256(unl).digest()).digest()
        checksum = checksum[0:4]
        unl = unl + checksum
        unl = base64.b64encode(unl)
        unl = unl.decode("utf-8")
        
        return unl

if __name__ == "__main__":
    # global direct_net
    # unl = UNL(direct_net)
    # print(unl.construct())

    # print(unl.deconstruct('x'))

    # print(unl.deconstruct(unl.construct()))
    pass





import requests

try:
    from urllib.parse import urlencode
except:
    from urllib import urlencode

import random

try:
    import json
except:
    import simplejson as json

import binascii

try:
    from Queue import Queue, Full  # py2
except ImportError:
    from queue import Queue, Full  # py3

dht_msg_endpoint = "http://158.69.201.105/pyp2p/dht_msg.php"

class DHT():
    def __init__(self):
        self.node_id = self.rand_str(20)
        self.password = self.rand_str(30)
        self.messages_received = Queue(maxsize=100)

        # Register a new "account."
        self.register(self.node_id, self.password)

    def rand_str(self, length):
        return ''.join(random.choice('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ') for i in range(length))

    def register(self, node_id, password):
        # Registers a new node to receive messages.
        call = dht_msg_endpoint + "?call=register&"
        call += urlencode({"node_id": node_id}) + "&"
        call += urlencode({"password": password})

        # Make API call.
        response = requests.get(call)

    def put(self, node_id, msg):
        # Send a message directly to a node in the "DHT"
        call = dht_msg_endpoint + "?call=put&"
        call += urlencode({"node_id": node_id}) + "&"
        call += urlencode({"msg": msg})

        # Make API call.
        response = requests.get(call)

    def list(self, node_id, password):
        # Get messages send to us in the "DHT"
        call = dht_msg_endpoint + "?call=list&"
        call += urlencode({"node_id": node_id}) + "&"
        call += urlencode({"password": password})

        # Make API call.
        response = requests.get(call).text
        response = json.loads(response)

        # Return a list of messages.
        return response

    def send_direct_message(self, node_id, msg):
        if type(node_id) != str:
            node_id = node_id.decode("utf-8")

        self.put(node_id, msg)

    def get_id(self):
        return self.node_id.encode("ascii")

    def has_messages(self):
        for msg in self.list(self.node_id, self.password):
            self.messages_received.put_nowait(msg)

        return not self.messages_received.empty()

    def get_messages(self):
        result = []
        while not self.messages_received.empty():
            result.append(self.messages_received.get())

        return result


if __name__ == "__main__":
    dht_node = DHT()

    dht_node.send_message(dht_node.node_id, u"sdfsdf")
    dht_node.send_message(dht_node.node_id, b"sdfsdf")
    print(dht_node.node_id)
    print(dht_node.password)
    print(dht_node.list(dht_node.node_id, dht_node.password))


